										;'This is a fake file

[test_relations]
PHA=BMI,AGE,CYCLL,DIAG,DISDUR,GEOREG,LBEST,MENS,WT
PHIA=BMI,AGE,CYCLL,DIAG,DISDUR,GEOREG,LBEST,MENS,WT
SLP=BMI,AGE,CYCLL,DIAG,DISDUR,GEOREG,LBEST,MENS,WT,CPL
PHIB=BMI,AGE,CYCLL,DIAG,DISDUR,GEOREG,LBEST,MENS,WT
